==> cbrkit-0.1.0-py3-none-any.whl: cbrkit/sim/taxonomy.py <==
from dataclasses import dataclass, field
from typing import Optional, Protocol, TypedDict, cast

from cbrkit.loaders import data as load_data
from cbrkit.typing import FilePath, SimPairFunc, SimVal


class SerializedNode(TypedDict, total=False):
    key: str
    sim: float
    children: list["SerializedNode | str"]


@dataclass
class TaxonomyNode:
    key: str
    weight: float | None
    depth: int
    parent: Optional["TaxonomyNode"]
    children: dict[str, "TaxonomyNode"] = field(default_factory=dict)


class Taxonomy:
    root: TaxonomyNode
    nodes: dict[str, TaxonomyNode]

    def __init__(self, path: FilePath) -> None:
        root_data = cast(SerializedNode, load_data(path))
        self.nodes = {}
        self.root = self._load(root_data)

    def _load(
        self,
        data: SerializedNode | str,
        parent: TaxonomyNode | None = None,
        depth: int = 0,
    ) -> TaxonomyNode:
        if isinstance(data, str):
            data = {"key": data}

        assert "key" in data, "Missing key in some node"

        node = TaxonomyNode(
            key=data["key"],
            weight=data.get("weight"),
            depth=depth,
            parent=parent,
        )

        for child in data.get("children", []):
            child_node = self._load(child, node, depth + 1)
            node.children[child_node.key] = child_node

        self.nodes[node.key] = node

        return node

    def lca(self, node1: TaxonomyNode, node2: TaxonomyNode) -> TaxonomyNode:
        while node1 != node2:
            if node1.parent is None or node2.parent is None:
                return self.root

            if node1.depth > node2.depth:
                node1 = node1.parent
            else:
                node2 = node2.parent

        return node1


class TaxonomyFunc(Protocol):
    def __call__(self, taxonomy: Taxonomy, x: str, y: str) -> SimVal:
        ...


def wu_palmer() -> TaxonomyFunc:
    def wrapped_func(taxonomy: Taxonomy, x: str, y: str) -> SimVal:
        node1 = taxonomy.nodes[x]
        node2 = taxonomy.nodes[y]
        lca = taxonomy.lca(node1, node2)

        return (2 * lca.depth) / (node1.depth + node2.depth)

    return wrapped_func


_taxonomy_func = wu_palmer()


def load(path: FilePath, measure: TaxonomyFunc = _taxonomy_func) -> SimPairFunc[str]:
    taxonomy = Taxonomy(path)

    def wrapped_func(x: str, y: str) -> SimVal:
        return measure(taxonomy, x, y)

    return wrapped_func

==> cbrkit-0.3.1-py3-none-any.whl: cbrkit/sim/taxonomy.py <==
from dataclasses import dataclass, field
from typing import Optional, Protocol, TypedDict, cast

from cbrkit.loaders import data as load_data
from cbrkit.typing import FilePath, SimPairFunc

__all__ = ["Taxonomy", "TaxonomyNode", "TaxonomyFunc", "load", "wu_palmer"]


class SerializedNode(TypedDict, total=False):
    key: str
    sim: float
    children: list["SerializedNode | str"]


@dataclass(slots=True)
class TaxonomyNode:
    key: str
    weight: float | None
    depth: int
    parent: Optional["TaxonomyNode"]
    children: dict[str, "TaxonomyNode"] = field(default_factory=dict)


class Taxonomy:
    __slots__ = ("root", "nodes")

    root: TaxonomyNode
    nodes: dict[str, TaxonomyNode]

    def __init__(self, path: FilePath) -> None:
        root_data = cast(SerializedNode, load_data(path))
        self.nodes = {}
        self.root = self._load(root_data)

    def _load(
        self,
        data: SerializedNode | str,
        parent: TaxonomyNode | None = None,
        depth: int = 0,
    ) -> TaxonomyNode:
        if isinstance(data, str):
            data = {"key": data}

        assert "key" in data, "Missing key in some node"

        node = TaxonomyNode(
            key=data["key"],
            weight=data.get("weight"),
            depth=depth,
            parent=parent,
        )

        for child in data.get("children", []):
            child_node = self._load(child, node, depth + 1)
            node.children[child_node.key] = child_node

        self.nodes[node.key] = node

        return node

    def lca(self, node1: TaxonomyNode, node2: TaxonomyNode) -> TaxonomyNode:
        while node1 != node2:
            if node1.parent is None or node2.parent is None:
                return self.root

            if node1.depth > node2.depth:
                node1 = node1.parent
            else:
                node2 = node2.parent

        return node1


class TaxonomyFunc(Protocol):
    def __call__(self, taxonomy: Taxonomy, x: str, y: str) -> float:
        ...


def wu_palmer() -> TaxonomyFunc:
    """Wu & Palmer similarity measure of two nodes in a taxonomy.
    >>> taxonomy = Taxonomy("./data/cars-taxonomy.yaml")
    >>> sim = wu_palmer()
    >>> sim(taxonomy, "audi", "porsche")
    0.5
    >>> sim(taxonomy, "audi", "bmw")
    0.0
    """

    def wrapped_func(taxonomy: Taxonomy, x: str, y: str) -> float:
        node1 = taxonomy.nodes[x]
        node2 = taxonomy.nodes[y]
        lca = taxonomy.lca(node1, node2)

        return (2 * lca.depth) / (node1.depth + node2.depth)

    return wrapped_func


_taxonomy_func = wu_palmer()


def load(
    path: FilePath, measure: TaxonomyFunc = _taxonomy_func
) -> SimPairFunc[str, float]:
    """Load a taxonomy and return a function that measures the similarity.
    >>> sim = load("./data/cars-taxonomy.yaml", measure=wu_palmer())
    >>> sim("audi", "porsche")
    0.5
    >>> sim("audi", "bmw")
    0.0
    """
    taxonomy = Taxonomy(path)

    def wrapped_func(x: str, y: str) -> float:
        return measure(taxonomy, x, y)

    return wrapped_func
